print.pseudorates <- function(x, ...) {
  cat("Call:\n")
  print.default(x$call)
  cat("\n", x$taxa, "taxa with", x$phyl, "gene families and", (length(x$w)), "different phyletic gene patterns.\n")
  cat("-----------------------------------\n")
  cat("Groups of nodes with the same rates:\n")
  print.default(x$bg)
  cat("-----------------------------------\n")
  cat("Parameter Index Matrix\n")
  print(x$modelmat)
  cat("-----------------------------------\n")
  for (i in x$modelnames) {
    cat(i, "\n")
    if (x$results[[i]]$convergence == 0) {
      print.default(x$results[[i]]$parsep)
      cat("Loglikelihood for model", i, ":", -x$results[[i]]$objective, "\n")
      cat("AIC           for model", i, ":", x$results[[i]]$AIC, "\n")
      cat("BIC           for model", i, ":", x$results[[i]]$BIC, "\n")
      cat("-----------------------------------\n")
    } else{
      cat("Convergence not achieved.\nIncrease number of iterations and/or function calls. See control arguments for the optimization method used.\n")
      cat("-----------------------------------\n")
    }
  }
  if (!all(x$conv == 0)) cat("Not all models converged. \n")
  cat("Time taken:", x$time[3], "seconds.\n")
}
plottree <- function(x, toilabel = TRUE, colors = NULL, ...) {
  if (is.null(colors))
    colors <- grDevices::rainbow(length(x$bg))
  colo <- NULL
  for (i in 1:nrow(x$tree$edge)) {
    tmp <- which(unlist(lapply(lapply(x$bg, FUN = function(X) c(x$tree$edge[i, 1], x$tree$edge[i, 2]) %in% X), FUN = "all"), use.names = FALSE))
    if ((length(tmp) > 1)) {
      colo[i] <- tmp[which.min(unlist(lapply(x$bg, FUN = length)))]
    } else{
      colo[i] <- tmp
    }
  }
  if (toilabel)
    x$tree$tip.label[x$toi] <- paste(x$tree$tip.label[x$toi], "+", sep = "")
  ape::plot.phylo(x$tree, edge.color = colors[colo], ...)
}
checkbglist <- function(bg, tree){
  all_branches <- TRUE
  for (i in 1:length(tree$edge.length)) {
    if (!any(unlist(lapply(lapply(bg, FUN = function(j) tree$edge[i,] %in% j),FUN = "all")))) all_branches <- FALSE
  }
  return(all_branches)
}
pseudorates <- function(usertree = NULL, userphyl = NULL, matchtipstodata = FALSE, unobserved = NULL, alphabet=NULL, modelmat = NULL, toi = 1, bgtype = "listofnodes", bg = NULL, numhessian=TRUE, rootprob = NULL, rpvec=NULL, lowli = 0.00001, upli = 100, ...) {

  ptm <- proc.time()
  bgtype <- match.arg(bgtype, c("listofnodes", "ancestornodes"))
  if (is.null(usertree))
    stop("usertree option is required.")
  if (is.null(userphyl) | !(is.matrix(userphyl) | is.data.frame(userphyl)))
    stop("userphyl option is required.")
  if (!is.null(unobserved)) {
    if (!(is.matrix(unobserved) | is.data.frame(unobserved))) {
      stop("unobserved option should be a matrix.")
    }
  }
  if (!is.null(unobserved)) {
    if (ncol(userphyl) != ncol(unobserved)) {
      stop("userphyl and unobserved should have the same number of taxa (columns).")
    }
  }
  if (is.null(alphabet))
    stop("alphabet option is required")
  if (is.null(modelmat))
    stop("modelmat option is required")
  if (nrow(modelmat) != ncol(modelmat))
    stop("A square matrix should be provided for the modelmat option.")
  if (nrow(modelmat) != length(alphabet))
    stop("Length of alphabet supplied does not equal the number of states in the supplied rate matrix.")
  if (bgtype == "listofnodes" & !is.null(bg) & is.list(bg)) {
    if (!checkbglist(bg, usertree)) stop("Some branch is missing in bg.")
  }
  if (nrow(userphyl) < ncol(userphyl))
    warning("More taxa than patterns detected. Note that rows should represent the discrete character patterns and columns the different taxa. If this format was used, you can safely ignore this message.")

  library("ape")
  library("numDeriv")
  library("Rcpp")
  library("markophylo")
  #########Transition rate matrix and substitution rate matrix#######
  TPM_taxa <- function(rates, ad, ti) {
    j  <- unlist(lapply(lapply(bg, FUN = function(X) c(ad[1], ad[2]) %in% X), FUN = "all"))
    for (gf in 1:length(rates[,j])) {
      Q[modelmat == gf] <- rates[gf,j]
    }
    diag(Q)   <- -rowSums(Q, na.rm = TRUE)
    return(ape::matexpo(Q*ti))
  }
  ###########Data Import#########
  if (ape::is.binary.tree(usertree) == FALSE | ape::is.rooted(usertree) == FALSE) {
    usertree <- ape::multi2di(usertree)
    cat("Tree either not binary or not rooted.","\n","Transformed using multi2di from ape.","\n")
    cat("Labels of nodes of interest might change.\n")
  }
  tree1 <- usertree
  tree1 <- ape::reorder.phylo(tree1, order = "postorder")
  if (is.data.frame(userphyl)) userphyl <- as.matrix(userphyl)
  if (is.data.frame(unobserved)) unobserved <- as.matrix(unobserved)
  if (matchtipstodata) {
    fin <- userphyl[, pmatch(usertree$tip.label, colnames(userphyl))]
    datab <- t(fin)
  } else {
    datab <- t(userphyl)
  }
  phyl  <- ncol(datab)
  nooftaxa <- length(tree1$tip.label)
  al <- length(alphabet)
  Q <- modelmat
  if (bgtype == "listofnodes" & is.null(bg)) {
    bg <- list(c(1:(nooftaxa * 2 - 1)))
  } else if (bgtype == "listofnodes" & !is.null(bg)) {
    bg <- bg
  } else if (bgtype == "ancestornodes") {
    uncl <- sort(bg, decreasing = TRUE) #unique clades
    for (i in 1:length(uncl)) {
      if (!is.na(uncl[i])) {
        try <- phangorn::Ancestors(tree1, uncl[i], type = "all")
        res <- match(try,uncl)
        uncl[na.omit(res)] <- NA
      }
    }
    bg <- uncl <- na.omit(uncl)
    ########
    if (is.null(bg))
      stop("bg not specified.")
    bgo <- bg
    bg <- list()
    for (i in 1:length(bgo)) bg[[i]] <- c(bgo[i], phangorn::Descendants(tree1, bgo[i], type = "all"))
    bg[[length(bg) + 1]] <- c(setdiff(1:(nooftaxa * 2 - 1), c(unlist(bg))), bgo)
  }
  ##############Parameters and Variables################
  nointnodes <- nooftaxa - 1
  tips <- 1:nooftaxa
  csp  <- length(bg) #how many different clade-specific parameters
  toi <- toi
  databp <- datab
  len_tips <- length(tips)


  pat <- markophylo::patterns(t(databp))
  w <- pat$w
  databp_red <- pat$databp_red

  logll <- NULL
  results <- list()
  nodelist <- c(setdiff(tree1$edge[, 2], tips), length(tips) + 1)
  F <- cbind(tree1$edge, tree1$edge.length)
  npar_mmat <- sort(unique(as.vector(modelmat[modelmat > 0])),na.last = NA)
  le_npar_mmat <- length(npar_mmat)
  rates <- matrix(NA, nrow = le_npar_mmat, ncol = csp)
  ###############Log-likelihood function###############
  bamsp <- function(previtval, model) { #branch and model specific assignment
    le_prev <- length(previtval)
    if (model == "wp") {
      rates[] <- previtval[1:(le_npar_mmat*csp)]
      p <- previtval[(le_npar_mmat*csp + 1):(le_npar_mmat*csp + 4)]
      return(list(rates = rates, p = p))
    }
  }
  update_Lixi_init <- function(miss_curr, datau, Lixiu){
    p <- miss_curr
    for (i in 1:nrow(datau)) {
      dta <- datau[i, ][toi] == "a"
      dtb <- datau[i, ][toi] == alphabet[1]
      dtc <- datau[i, ][toi] == alphabet[2]
      dtd <- datau[i, ][toi] == alphabet[3]
      ps <- p
      if (length(toi) > 1) {
        Lixiu[[i]][toi, ][dta, ] <- cbind(ps,  ps, ps ^ 2)[dta, ]
        Lixiu[[i]][toi, ][dtb, ] <- cbind(1 - ps,  0,
                                          ps * (1 - ps))[dtb, ]
        Lixiu[[i]][toi, ][dtc, ] <- cbind(0,  1 - ps,
                                          ps * (1 - ps))[dtc, ]
        Lixiu[[i]][toi, ][dtd, ] <- cbind(0,  0,
                                          (1 - ps)*(1 - ps))[dtd, ]
      }
    }
    return(Lixiu)
  }

  tmp <- numeric(al)

  alseq <- 1:length(alphabet)


  ll <- function(model, pm, rootp, Lixi_in_ll) {
    tmp <- allpatt_loopC(nodelist, al, tree1$edge[, 1], tree1$edge[, 2], pm, Lixi_in_ll, len_tips)
    logll <- log(rowSums(sweep(tmp, MARGIN = 2, rootp, `*`)))
    return(logll)
  }
  #################Objective function################
  totalll <- function(previtval, model, Lixi_in_totll, Lixi_corrin_totll) {
    ptbrun <- bamsp(previtval, model)
    ptb_loc <- ptbrun$rates
    Lixi_ob <- update_Lixi_init(ptbrun$p, databp_red, Lixi_in_totll)
    Lixi_unob <- update_Lixi_init(ptbrun$p, unobserved, Lixi_corrin_totll)
    rootp <- rpvec

    pm  <- lapply(1:nrow(F), function(i) TPM_taxa(ptb_loc, ad = F[i, 1:2], ti = F[i, 3]))
    phy <- w * ll(model, pm, rootp, Lixi_in_ll = Lixi_ob)
    if (!is.null(unobserved)) {
      corr <- ll(model, pm, rootp, Lixi_in_ll = Lixi_unob)
      return(-(sum(phy) - ncol(databp) * log(1 - sum(exp(corr))) ) )
    } else {
      return(-(sum(phy)))
    }
    #-ve of the value because we are minimizing by default.
  }
  indelinit <- 0.9
  missingguess <- rep(0.05, 1)
  modelnames = c("wp")
  ################Estimation################
  options(digits = 7)
  modelop <- list(wp = list(start = c(rep(indelinit, (csp*le_npar_mmat)), rep(missingguess, 4)), #rate pars, p
                            df = 1 * csp * le_npar_mmat + 1 * 4, pb = 2, lower = c(rep(c(lowli), (csp*le_npar_mmat)), rep(0, 4)),
                            upper = c(rep(c(upli),
                                          (csp*le_npar_mmat)), rep(0.99, 4)),
                            model = "wp") )
  parvec <- NULL
  sevec <- NULL
  convcheck <- NULL

  for (i in modelnames) {
    Lixi_init <- rep(list(matrix(data = 0, nrow = nooftaxa + nointnodes, ncol = al)), length = length(w))
    for (k in 1:length(w)) {
      for (u in alseq) {
        Lixi_init[[k]][which(databp_red[k, ] == alphabet[u]), u] <- 1
      }
    }
    Lixi_corr_init <- rep(list(matrix(data = 0, nrow = nooftaxa + nointnodes, ncol = al)), length = nrow(unobserved))
    for (k in 1:nrow(unobserved)) {
      for (u in alseq) {
        Lixi_corr_init[[k]][which(unobserved[k, ] == alphabet[u]), u] <- 1
      }
    }
    res <- nlminb(start = modelop[[i]]$start, objective = totalll, model = i, lower = modelop[[i]]$lower, upper = modelop[[i]]$upper, Lixi_in_totll = Lixi_init, Lixi_corrin_totll = Lixi_corr_init, ...)
    if (numhessian) res$hessian <- numDeriv::hessian(func = totalll, model = i, x = res$par, Lixi_in_totll = Lixi_init, Lixi_corrin_totll = Lixi_corr_init)
    res$parsep  <- bamsp(res$par, i)
    res$df <- modelop[[i]]$df
    convcheck <- c(convcheck, res$convergence)
    if (res$convergence == 0) {
      estimate <- res$par
      parvec   <- c(parvec, res$par[1:modelop[[i]]$pb])
      if (numhessian) {
        res$se   <- sqrt(diag(solve(res$hessian)))
        res$parsep$se <- bamsp(res$se, i)
        sevec <- c(sevec, res$se)
      }
      llhat <- -round(res$objective, 3)
      res$AIC <- 2 * llhat - 2 * res$df
      res$BIC <- 2 * llhat - log(phyl) * res$df
    }
    results[[i]] <- res
    if (res$convergence == 0) {
    } else{
      cat("Convergence not achieved.\nIncrease number of iterations and/or function calls. See control arguments for the optimization method used.\n")
    }
  }
  if (lowli %in% parvec || upli %in% parvec) {
    cat("Estimated parameters on interval bounds.")
  }
  if (!all(is.finite(sevec)))
    cat("Something is not right with the standard errors. Check Hessian matrix estimate. Try bootstrapping and use option numhessian = FALSE.")

  timetaken <- proc.time() - ptm
  val <- list(call = match.call(), conv = convcheck, time = timetaken, bgtype = bgtype, bg = bg, results = results, tree = tree1, data_red = databp_red, w = w, missingguess = missingguess, toi = toi,
              lowli = lowli, upli = upli, taxa = nooftaxa, phyl = ncol(databp), rootprob = rootprob, modelnames = modelnames, modelmat = modelmat)
  class(val) <- "pseudorates"
  return(invisible(val))
}
